/*
	LL-test - simple Lucas-Lehmer merssenne prime tester

	Copyleft Rune Magnussen 2008,2009

	22-05-2009: Wrong results when testing very small exponents
	            have been corrected.
		    Factoring resticted to f = 2kp+1

	30-09-2009: Added progres-indicator for exponents > 1000

	16-10-2009: Added likely/unlikely macros.

	08-11-2009: Factoring less verbose.
*/

#include <stdlib.h> /* strtol(), exit() */
#include <unistd.h> /* nice() */
#include <stdio.h>
#include <gmp.h>

#define SILENT_LIMIT 1000
#define SILENT_FACT_LIMIT 20000
#define NICE_ADD 10

/* Stolen fom NASM sources */
#if defined(__GNIC__) && __GNUC__ >= 3
# define likely(x)	__builtin_expect(!!(x), 1)
# define unlikely(x)	__builtin_expect(!!(x), 0)
#else
# define likely(x) (!!(x))
# define unlikely(x) (!!(x))
#endif

/* #define DEBUG 1 */

#ifdef DEBUG
#define dprintf(fmt, a, b) printf(fmt, a, b)
#else
#define dprintf(fmt, a, b)
#endif

/*
Return 1 if n is prime and 0 otherwise.
*/
static int is_prime(const long n,int print) {
	long i;
	mpz_t max;
	long stop;
	int res;
	if (unlikely(n == 2))
		return 1;
	if (unlikely((n%2 == 0) || (n < 2)))
		return 0;
	mpz_init_set_ui(max,n);
	mpz_sqrt(max,max);
	mpz_add_ui(max,max,1L);
	stop = mpz_get_si(max);
	dprintf("Factoring %ld to %ld\n",n,mpz_get_ui(max));
	mpz_clear(max);
	for (i=3,res=1;i<stop;i+=2) {
		if (n % i == 0) {
			if (unlikely(print))
				printf("Factor found: %ld\n",i);
			res = 0;
			break;
		}
	}
	return res;
}

/*
Return 1 if a factor is found, 0 otherwise.
Even if 0 is returned m may be composite.

Factors have the form f = 2*e*k+1.
and are tested if f = 1 or 7 mod 8
*/
static int is_composite(mpz_t *m, const unsigned long e) {
	int res = 0, v = (e > SILENT_FACT_LIMIT);
	unsigned long i,f;
	unsigned long e2 = e << 1,trig = e2 / 100;
	if (unlikely(mpz_get_ui(*m)<e2))
		return 0;
	if (unlikely(trig < 1))
		trig = 1;
	dprintf("Looking for factor between %ld and %ld\n",e2+1,e2*(e2-1)+1);
	for (i=0;i<e2;i++) {
		f = e2*i+1;
		if (unlikely(f>=mpz_get_ui(*m)))
			return 0;
		if (((f % 8 == 1) || (f % 8 == 7)) && is_prime(f,0)) {
			dprintf("Testing %ld in iter %ld\n",f,i);
			if (mpz_divisible_ui_p(*m,f)) {
				if (unlikely(v))
					printf("\n");
				printf("Factor found: %ld\n",f);
				res = 1;
				break;
			}
		}
		if (unlikely(v && i % trig == 0)) {
			printf(".");
			fflush(NULL);
		}
	}
	if (unlikely(v))
		printf("\n");
	return res;
}

/*
Perform lltest unless a factor is found
*/
static long lltest(const unsigned long e) {
	long res = 0;
	unsigned long i;
	unsigned long max = e-2;
	/* 2 is a special case */
	if (unlikely(!max))
		return 0;

/*#ifdef DEBUG */
	unsigned long trig;
	trig = max/100;
	if (unlikely(!trig)) {
		trig = 1;
	}
/*#endif */
	mpz_t m, s, t;
	mpz_init_set_ui(m,2L);	
	mpz_init_set_ui(t,1L);
	mpz_mul_2exp(m,t,e);
	mpz_clear(t);
	mpz_sub_ui(m,m,1L);
	if (likely(is_composite(&m,e))) {
		mpz_clear(m);
		return 1; /* a prime results in 0 */
	}
	printf("Starting Lucas-Lehmer test of M(%ld)\n",e);
	mpz_init_set_ui(s,4L);
	for (i=0;i<max;++i) {
		mpz_powm_ui(s,s,2L,m); /* pow and mod */
		mpz_sub_ui(s,s,2L);
/* #ifdef DEBUG */
		if (unlikely(e>SILENT_LIMIT && i % trig == 0 && i > 0)) {
		/*	
			printf("Iteration %ld of %ld\n",i,max);
		*/
			printf(".");
			if (i % (trig*20) == 0) {
				printf("\n");
			} else {
				fflush(NULL);
			}
		}
/* #endif */
	}
	printf("\n");
	res = mpz_get_si(s);
	dprintf("s:%c%ld\n",' ',res);
	mpz_clear(s);
	mpz_clear(m);
	return res;
}

static void usage(char* pgm) {
	fprintf(stderr,"Usage: %s positive integer\n",pgm);
	exit(EXIT_FAILURE);
}

int main(int argc,char *argv[]) {
	signed long exp; /* Signed in case of negative input */
	if (likely(argc == 2)) {
		exp = strtol(argv[1],NULL,10);
		if (likely(exp>0)) {
			nice(NICE_ADD);
			if (is_prime(exp,1)) {
				if(likely(lltest(exp))) {
					printf("M(%s) is not prime.\n",argv[1]);
				} else {
					printf("M(%s) is prime!\n",argv[1]);
				}
			} else {
				printf("%s is not prime.\n",argv[1]);
			}
		} else {
			usage(argv[0]);
		}			
	} else {
		usage(argv[0]);
	}
	return EXIT_SUCCESS;
}
